#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
from abc import ABC, abstractmethod
from typing import Any
import pandas as pd
import pyarrow as pa
from getfactormodels.utils.http_client import HttpClient


class FactorModel(ABC):
    """base model used by all factor models."""
    @property
    @abstractmethod
    def _frequencies(self) -> list[str]:
        pass

    @abstractmethod
    def _get_url(self) -> str:
        """builds the unique data source URL."""
        pass

    @abstractmethod
    def _read(self, data: bytes) -> pd.DataFrame | pa.Table:
        """convert bytes into a pd.DataFrame or pa.Table."""
        pass

    def __init__(self, frequency: str = 'm',
                 start_date: str | None  = None,
                 end_date: str | None = None,
                 output_file: str | None = None,
                 cache_ttl: int = 86400,
                 **kwargs: Any):

        logger_name = f"{self.__module__}.{self.__class__.__name__}"
        self.log = logging.getLogger(logger_name)

        self.frequency = frequency.lower()
        self.start_date = start_date
        self.end_date = end_date
        self.output_file = output_file
        self.cache_ttl = cache_ttl
        self._data: pd.DataFrame | None = None # Internal storage for processed data

        if self.frequency not in self._frequencies:
            raise ValueError(f"Invalid frequency {frequency}. Valid options: {self._frequencies}")
        super().__init__()

    @property
    def data(self) -> pd.DataFrame:
        """public: access the data. Checks for data, calls download()."""
        # Check if data is stored
        if self._data is not None:
            return self._data 
        # Download, not stored
        return self._download()


    def download(self) -> pd.DataFrame:
        """Public method to download and return the data. 
        Calls .data property to fetch and store
        """
        return self.data


    def extract(self, factor: str | list[str]) -> pd.Series | pd.DataFrame:
        """Retrieves a single factor (column) from the dataset."""
        data = self.download()

        if data.empty:
            self.log.error("DataFrame is empty.")
            raise RuntimeError("DataFrame empty: can not extract a factor.")

        if isinstance(factor, str):
            if factor not in data.columns:
                available = list(data.columns)
                self.log.error(f"Factor '{factor}' not found in model. Available: {available}")
                raise ValueError(f"Factor '{factor}' not available.")       
            return data[factor]

        elif isinstance(factor, list):
            return data[factor] 


    # TODO: Remove FactorExtractor
    #def _drop_rf():
        #if rf=0
    #def _drop_mktrf(): 
        #if mktrf = 0
    # set flags.
    #let utils handle it. (_pd_rearrange_cols, if no_rf = 1, then drop it)


    # Making download concrete, and moved the abstractmethod to _read!
    def _download(self) -> pd.DataFrame:
        """Private template method: called by `data` property when self._data is none.
        * Should not be called directly with subclasses.
        """
        # Don't need to check for data here
        raw_data = self._download_from_url()
        
        data = self._read(raw_data)

        if isinstance(data, pa.Table):
            data = data.to_pandas()
  
        # Storage
        self._data = data

        return data


    @property
    def _url(self) -> str:
        """Internal property: data source URL.
        Fama French returns the underlying model url (4 factor (3+MOM) returns the 3 factor url)
        """
        # subclasses implement _get_url which this uses.
        return self._get_url()

    def _download_from_url(self) -> bytes:
        url = self._url
        log_msg = f"Downloading data from: {url}"
        self.log.info(log_msg)
        try:
            with HttpClient(timeout=15.0) as client:
                return client.download(url, self.cache_ttl)
        except Exception as e:
            self.log.error(f"Failed to download from {url}: {e}")
            # crash fast on download failure
            raise RuntimeError(f"Download failed for {url}.") from e




#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
from typing import Any
import pandas as pd
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.models.fama_french import FamaFrenchFactors
from getfactormodels.utils.http_client import HttpClient
from getfactormodels.utils.utils import _process


class DHSFactors(FactorModel):
    # roughing in infos, not approp for docstr but need TODO a reliable
    # way of getting and setting these when more models are redone. Most
    # importantly the copyright/attribution info! TODO
    """Download the DHS Behavioural Factors.

    Downloads the Behavioural Factors of Kent Daniel, David Hirshleifer, and 
    Lin Sun (DHS). Data from 1972-07-01 until end of 2023.

    Args:
        `frequency` (`str`): The frequency of the data. `m` or `d` (default: `m`)
        `start_date` (`str, optional`): The start date of the data, `YYYY-MM-DD`
        `end_date` (`str, optional`): The end date of the data, `YYYY-MM-DD`
        `output_file` (`str, optional`): The filepath to save the output data.

    References:
    - Short and Long Horizon Behavioral Factors," Kent Daniel, David 
    Hirshleifer and Lin Sun, Review of Financial Studies, 2020, 33 (4):
    1673-1736.
    
    Data source: https://sites.google.com/view/linsunhome/
    """
    @property
    def _frequencies(self) -> list[str]:
        return ['d', 'm']

    SCHEMA = pa.schema([
        ('Date', pa.timestamp('ms')),
        ('FIN', pa.float64()),
        ('PEAD', pa.float64()),
    ])

    def __init__(self, frequency: str = 'm', **kwargs: Any) -> None:
        self.frequency = frequency
        super().__init__(frequency=frequency, **kwargs)


    def _get_url(self) -> str:
        """Construct the Google Sheet URL for monthly or daily."""
        base_url = 'https://docs.google.com/spreadsheets/d/'

        if self.frequency == 'd':
            gsheet_id = '1lWaNCuHeOE-nYlB7GA1Z2-QQa3Gt8UJC'
            #info_id =
        else:
            gsheet_id = '1VwQcowFb5c0x3-0sQVf1RfIcUpetHK46'
            #info_sheet_id = '#gid=96292754'

        return  f'{base_url}{gsheet_id}/export?format=csv'


    def _read(self, data):
        """Reads the Mispricing factors CSV."""
        try:
            if self.frequency == 'd':
                col_names = ['Date', 'Year', 'Month', 'Day', 'FIN', 'PEAD']
            else:
                col_names = ['Date', 'PEAD', 'FIN'] #note FIN PEAD swap in m and d 
            # TEST: reading CSV with PyArrow
            # will be a CSV reader
            read_opts = pv.ReadOptions(skip_rows=1, 
                                       column_names=col_names)

            parse_opts = pv.ParseOptions(delimiter=',',
                                         ignore_empty_lines=True)

            convert_opts = pv.ConvertOptions(
                column_types=self.SCHEMA,
                timestamp_parsers=["%m/%d/%Y", "%Y%m"],
                include_columns=['Date', 'PEAD', 'FIN'], 
            )

            table = pv.read_csv(
                io.BytesIO(data),
                read_options=read_opts,
                parse_options=parse_opts,
                convert_options=convert_opts,
            )

            table = table.rename_columns(['date', 'PEAD', 'FIN'])

            # for debug, testing 
            initial_rows = table.num_rows
            table = table.drop_null()

            final_rows = table.num_rows
            rows_dropped = initial_rows - final_rows

            if rows_dropped > 0:
                msg = ("{rows_dropped} NaN rows dropped." # check
                    f"{initial_rows} -> {final_rows}).")
                self.log.debug(msg)
            else:
                msg = f"No NaNs detected, {final_rows} rows."
                self.log.debug(msg)


            # pandas line --------------------------------------------------- #
            df = table.to_pandas()  # LOSS OF DATA/ROUNDING ERROR TODO (sheet has more than 4 decimals)

            df = df.set_index('date')
            df.index.name = 'date' # Set the index name
            
            data = df * 0.01    # TODO: Decimal types possibly
            data = data.round(8) 
            # TODO: check every model, does it need decimalization?

            if self.frequency == "m":
                data.index = data.index + pd.offsets.MonthEnd(0) 
                # bad TODO, start of month, THEN get data THEN offset. TODO: After fama french

            # Move this into a function
            try:
                ffdata = FamaFrenchFactors(model="3", frequency=self.frequency,
                                           start_date=self.start_date, end_date=self.end_date)
                ff = ffdata.download()
                #ff = ff.round(4)
                data = pd.concat([ff["Mkt-RF"], data, ff["RF"]], axis=1)
                data = data.dropna(how="any")

            except NameError:
                print("Warning: _get_ff_factors function not found. Skipping FF merge.")

            return _process(data, self.start_date,
                            self.end_date, filepath=self.output_file)

        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            self.log.error("Reading or parsing failed: %s", e)
            empty_df = pd.DataFrame(columns=['Date', 'PEAD', 'FIN'], # type: ignore [reportArgumentType] 
                                    index=pd.DatetimeIndex([], name='date'))
            return empty_df



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .fama_french import FamaFrenchFactors


class CarhartFactors(FamaFrenchFactors): # inheritence ooo
    """Download the Carhart 4-Factor model.
    
    Args:
        `frequency` (`str`): The data frequency. `d`, `m`, `y` (default: `m`).
        `start_date` (`str, optional`): The start date `YYYY-MM-DD`
        `end_date` (`str, optional`): The end date `YYYY-MM-DD`
        `output_file` (`str, optional`): Optional file path to save to file.
        `cache_ttl` (`int, optional`): Cached download time-to-live in seconds
            (default: `86400`).

    Returns:
        `pd.DataFrame`: timeseries of factors.

    References:
    - M. Carhart, ‘On Persistence in Mutual Fund Performance’, Journal
    of Finance, vol. 52, no. 1, pp. 57–82, 1997.

    Data source: https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "m", "y"]

    def __init__(self, frequency='m', **kwargs):
        model = '4'                     # enforce model 
        super().__init__(frequency=frequency, 
                         model=model,   # give it to ff 
                         **kwargs)



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
from typing import Any
import pandas as pd
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.utils import _process


class ICRFactors(FactorModel):
    """Download the Intermediary Capital Ratio of He, Kelly, Manela (2017)

    Args:
        frequency (str): The data frequency ('d', 'm', 'q') (default: 'm').
        start_date (str, optional): The start date YYYY-MM-DD for the factor data
        end_date (str, optional): The end date YYYY-MM-DD
        output_file (str, optional): optional filepath to save data to. Supports .csv, .pkl.
        cache_ttl (int): Time-to-live for cache in seconds (default: 86400).
    
    Returns:
        pd.DataFrame
    
    References:
    - Z. He, B. Kelly, and A. Manela, ‘Intermediary asset pricing: New evidence
    from many asset classes’, Journal of Financial Economics, vol. 126, no. 1, 
    pp. 1–35, 2017. (https://doi.org/10.1016/j.jfineco.2017.08.002)
    
    Data source: https://zhiguohe.net
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "m", "q"]

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _get_url(self) -> str:
        _file = {"d": "daily", 
                 "m": "monthly", 
                 "q": "quarterly"}.get(self.frequency)
        return f"https://zhiguohe.net/wp-content/uploads/2025/07/He_Kelly_Manela_Factors_{_file}_250627.csv"


    def _read(self, data) -> pd.DataFrame:
        """Helper to remove TODO"""
        _data = io.StringIO(data.decode('utf-8'))
       
        # back to pd, old func stuff
        df = pd.read_csv(_data)
        df = df.rename(columns={df.columns[0]: "date"})

        # TODO: moving similar date validations to base
        if self.frequency == "q":
            # Quarterly dates are in a YYYYQ format [19752 to 1975Q2]
            df["date"] = df["date"].astype(str)
            df["date"] = df["date"].str[:-1] + "Q" + df["date"].str[-1]
            # Converts YYYYQ to a timestamp at the eoq
            df["date"] = pd.PeriodIndex(df["date"], freq="Q").to_timestamp() \
                + pd.offsets.QuarterEnd(0)
#PATTERN
        elif self.frequency == "m":
            df["date"] = pd.to_datetime(df["date"], format="%Y%m")
            df["date"] = df["date"] + pd.offsets.MonthEnd(0)

        elif self.frequency == "d":
            df["date"] = pd.to_datetime(df["date"], format="%Y%m%d")

        df = df.rename(columns={
                "intermediary_capital_ratio": "IC_RATIO",
                "intermediary_capital_risk_factor": "IC_RISK_FACTOR",
                "intermediary_leverage_ratio_squared": "INT_LEV_RATIO_SQ",
                "intermediary_value_weighted_investment_return": "INT_VW_ROI",
            })

        df = df.set_index("date")
        
        # TODO: daily returns 15 decimals; 
        # Decimal type possibly to avoid rounding errors.
        df = df.round(8) if self.frequency == 'd' else df.round(4) # RUF/SIM108
        
        #TODO: check if decimalizing anything?

        return _process(df, self.start_date,
                        self.end_date, filepath=self.output_file)




# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
from typing import Any
import pandas as pd
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.utils import _process


class LiquidityFactors(FactorModel):
    """Download the Pastor-Stambaugh Liquidity Factors.

    * Only available in monthly data.

    Args:
        frequency (str): The data frequency, 'm'.
        start_date (str, optional): The start date YYYY-MM-DD.
        end_date (str, optional): The end date YYYY-MM-DD.
        output_file (str, optional): Optional file path to save to file. Supports csv, pkl.
        cache_ttl (int, optional): Cached download time-to-live in seconds (default: 86400).
    
    Returns:
        pd.Dataframe: timeseries of factors.

    References:
    - L. Pastor and R. Stambaugh, ‘Liquidity Risk and Expected Stock Returns’, 
      Journal of Political Economy, vol. 111, no. 3, pp. 642–685, 2003.
    
    Data source: https://finance.wharton.upenn.edu/~stambaug/
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["m"]

    def __init__(self, frequency: str = 'm', **kwargs: Any) -> None:
        super().__init__(frequency=frequency, **kwargs)
 
    def _get_url(self) -> str:
        #TODO: Backup data sources: https://research.chicagobooth.edu/-/media/research/famamiller/data/liq_data_1962_2024.txt')
        return 'https://finance.wharton.upenn.edu/~stambaug/liq_data_1962_2024.txt'


    # Still old func stuff below here. Need to move some to base, but when consistent across models. 
    def _read(self, data) -> pd.DataFrame:
        _data = data.decode('utf-8')
        data = io.StringIO(_data)

        # Note: headers are the last commented line in header.
        headers = [line[1:].strip().split('\t')
            for line in data.readlines() if line.startswith('%')][-1]

        # Fix: was losing first line of data
        data.seek(0)
        data = pd.read_csv(data, sep='\\s+',
                           names=headers, 
                           comment='%',
                           index_col=0, 
                           na_filter=True, 
                           na_values=[-99, '-99.000000'], # we returning NaNs now!  #TODO: test to check the first x values 
                           )

        data.index.name = 'date'  # Should make it all DATE
        data.index = data.index.astype(str)

        data = data.rename(columns={'Agg Liq.': 'AGG_LIQ',
                                    'Innov Liq (eq8)': 'INNOV_LIQ',
                                    'Traded Liq (LIQ_V)': 'TRADED_LIQ'})

        data.index = pd.to_datetime(data.index, 
                                    format='%Y%m') + pd.offsets.MonthEnd(0)  

        # some things...
        data = data.round(6)
        # Need to either consistently return start or end of month across all 
        #  models, and decide whether or not to use BDay start/end (depending 
        #  on what the source data does. Don't think any models have a 
        #  weekend/non bday? maybe?..)

        # Check for nans, warn nd return 

        return _process(data, self.start_date,
                        self.end_date, filepath=self.output_file)



# PLACEHOLDER
#
# will be _Writer class.


# from utils


# TODO: Will redo as a Writer class with use pyarrow
# changing: no longer uses filename, output_dir, just filepath. Always returns Path 
def _prepare_filepath(filepath=None) -> Path:
    if filepath is None:
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        filepath = Path.cwd() / f"data_{timestamp}.csv"
        print(f"No filepath provided, creating: {filepath.name}")
        return filepath
    
    filepath = Path(filepath).expanduser()
    
    if filepath.is_dir():
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        filepath = filepath / f"data_{timestamp}.csv"
        print(f"Directory provided, creating: {filepath.name}")
    
    return filepath

def _save_to_file(data, filepath=None):
    if not isinstance(data, (pd.DataFrame, pd.Series)):
            raise ValueError('Data is not a pandas DataFrame or Series')
    full_path = _prepare_filepath(filepath)
    
    if full_path is None:
        raise ValueError("Failed to prepare filepath")
    
    extension = full_path.suffix.lower()
 
    if full_path.is_file():
        print(f'File exists: {full_path.name} - overwriting...')

    try:
        if extension == '.txt':
            data.to_csv(str(full_path), sep='\t')
        elif extension == '.csv':
            data.to_csv(str(full_path))
        elif extension == '.xlsx':
            data.to_excel(str(full_path))
        elif extension == '.pkl':
            data.to_pickle(str(full_path))
       # elif extension == '.md':
       #    #.md removed for now 
        else:
            supported = ['.txt', '.csv', '.xlsx', '.pkl']
            raise ValueError(f'Unsupported file extension: {extension}. Must be one of: {supported}')

        print(f"File saved to: {full_path}")
    except Exception as e:
        raise IOError(f"Failed to save file to {full_path}: {str(e)}")



#!/usr/bin/env python3
import logging
from pathlib import Path
from typing import Optional
from diskcache import Cache

logging.basicConfig(level=logging.ERROR)
log = logging.getLogger(__name__)

class _Cache:
    def __init__(self, cache_dir: str, default_timeout=86400):

        cache_path = Path(cache_dir) 
        
        self.cache = Cache(str(cache_path))

        self.default_timeout = default_timeout

        msg = f"Cache initialized: {cache_path}"
        log.debug(msg)

    @property
    def directory(self) -> str:
        return self.cache.directory

    def get(self, key: str) -> Optional[bytes]: #type err as bytes|None TODO FXME
        """Retrieves data from cache if valid, otherwise returns None."""
        data = self.cache.get(key)
        msg = f"key: {key}"
        log.debug(msg)
        if data is not None:
            log.debug("hit")
        else:
            log.debug("miss")
        return data

    def set(self, key: str, data: bytes, expire_secs: int | None = None):
        """Save data to the cache expiration time."""
        timeout = expire_secs if expire_secs is not None else self.default_timeout
        try:
            self.cache.set(key, data, expire=timeout)
            log.debug(f"data written to cache ({key[:8]}...): expiry: {timeout}s")
        except Exception as e:
            msg = f"Failed to write to cache: {key}:\n{e}"
            log.error(msg)

    def close(self):
        self.cache.close()
        msg = f'closed {self.cache.__class__.__name__}'
        log.debug(msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def clear(self):
        """Removes all data entries from the cache directory."""
        self.cache.clear()
        
        msg = f"cache cleared: (dir: {self.directory})"
        log.info(msg)



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
from typing import Any
import pandas as pd
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.utils import _process


class QFactors(FactorModel):
    """Download q-factor data from global-q.org.

    Args:
        `frequency` (`str`): the frequency of the data. `d`, `m` `y` or `w`
            (default: `m`).
        `start_date` (`str, optional`): the start date of the data, `YYYY-MM-DD`
        `end_date` (`str, optional`): the end date of the data, as `YYYY-MM-DD`
        `classic` (`bool, optional`): returns the original 4-factor q-factor 
            model. (Default: `False`).
        `cache_ttl` (`int, optional`): Cached download time-to-live in seconds 
            (default: `86400`).
    Returns:
        `pd.DataFrame`: timeseries of factors

    References:
    - Hou, Kewei, Haitao Mo, Chen Xue, and Lu Zhang, 2021, An augmented q-factor model
    with expected growth, Review of Finance 25 (1), 1-41. (q5 model)
    - Hou, Kewei, Chen Xue, and Lu Zhang, 2015, Digesting anomalies: An investment
    approach, Review of Financial Studies 28 (3), 650-705. (Classic q-factor model)

    Data Source: https://global-q.org/factors.html
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "w", "w2w", "m", "q", "y"] # test

    def __init__(self, *, classic: bool = False, **kwargs: Any) -> None:
        self.classic = classic 
        super().__init__(classic=classic, **kwargs)

    def _get_url(self) -> str:
        file = {'m': "monthly", 
                "d": "daily",
                "q": "quarterly", 
                "w": "weekly",
                "w2w": "weekly_w2w",
                "y": "annual"
                }.get(self.frequency)

        url = 'https://global-q.org/uploads/1/2/2/6/122679606'
        url += f'/q5_factors_{file}_2024.csv' # TODO: YEAR
        return url

    def _read(self, data) -> pd.DataFrame:
        _file = io.StringIO(data.decode('utf-8'))
        index_cols = [0, 1] if self.frequency in ["m", "q"] else [0]
        data = pd.read_csv(_file, parse_dates=False, index_col=index_cols, float_precision="high")

        return self._parse_q_factors(data)


    # ------------------------------------------------------------------ #
    def _parse_q_factors(self, data) -> pd.DataFrame:
        if self.frequency in ["m", "q"]: 
            data = data.reset_index()

            # Need to insert "-" or "Q" into date monthly/quarterly str.
            col = "quarter" if self.frequency == "q" else "month"
            char = "q" if self.frequency == "q" else "-"

            # Combines year, period cols into a PeriodIndex to Timestamp
            data["date"] = pd.PeriodIndex(
                data["year"].astype(str) + char + data[col].astype(str),
                freq=self.frequency.upper(),  #fix:FutureWarning 'm' is deprecated, use 'M'
            ).to_timestamp(how="end")

            data["date"] = data["date"].dt.normalize()

            data = data.drop(["year", col], axis=1).set_index("date")

        elif self.frequency == "y":
            data.index = pd.to_datetime(data.index.astype(str)) + pd.offsets.YearEnd(0)

        else: # daily, weekly
            data.index = pd.to_datetime(data.index.astype(str))

        data.columns = data.columns.str.upper()
        data.index.name = "date"

        if self.classic:
            data = data.drop(columns=["R_EG"])

        # TODO: check: its market, not mkt-rf!
        data = data.rename(columns={"R_MKT": "Mkt-RF", "R_F": "RF"})

        data = data * 0.01

        #TODO: slicing, saving, validating date to base model? util? Use here.
        return _process(data, self.start_date, self.end_date,
                        filepath=self.output_file)



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# TODO: break this all up into models/*.py !
# TODO: httpx, a client class, model classes... (done done and done)
import io
from typing import Any
import pandas as pd
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.utils import _process


class HMLDevilFactors(FactorModel): # Note HMLDevil -> HMLDevilFactors (to keep consistent)
    """Retrieve the HML Devil factors from AQR.com.

    Notes:
    - Very slow. This model implements a cache, and second run today/this month 
      should be faster.

    Parameters:
        frequency (str): The frequency of the data. M, D (default: M)
        start_date (str, optional): The start date of the data, YYYY-MM-DD.
        end_date (str, optional): The end date of the data, YYYY-MM-DD.
        output_file (str, optional): The filepath to save the output data.

    Returns:
        pd.DataFrame: the HML Devil model data indexed by date.
    
    References:
        C. Asness and A. Frazzini, ‘The Devil in HML’s Details’, The Journal of Portfolio Management, vol. 39, pp. 49–68, 2013.
    Data source: https//aqr.com
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "m"]  # TODO: Check AQR's HML Devil frequencies again. 

    def __init__(self, frequency: str = 'm', cache_ttl: int = 43200, **kwargs: Any) -> None:
        self.cache_ttl = cache_ttl
        super().__init__(frequency=frequency, cache_ttl=cache_ttl, **kwargs)

    def _get_url(self) -> str:
        base_url = 'https://www.aqr.com/-/media/AQR/Documents/Insights/'
        file = 'daily' if self.frequency == 'd' else 'monthly'

        return f'{base_url}/Data-Sets/The-Devil-in-HMLs-Details-Factors-{file}.xlsx'

# --------------------------------------------------------------------------------#
    def _read(self, data) -> pd.DataFrame:
        """Processes the downloaded Excel file from AQR."""
        sheets = {0: 'HML Devil', 4: 'MKT', 5: 'SMB', 7: 'UMD', 8: 'RF'}
        
        df_dict = pd.read_excel(
            io.BytesIO(data),
            sheet_name=list(sheets.values()),
            skiprows=18,
            header=0,
            engine='calamine', # The right choice for speed
            index_col=0,
            parse_dates=True,
        )

        dfs = []
        for sheet_index, sheet_name in sheets.items():
            df = df_dict[sheet_name]
            
            if sheet_index in [8, 0]:
                # RF and HML Devil (iloc[:, 0:1])
                df_slice = df.iloc[:, 0:1]
            else:
                # MKT, SMB, UMD (uses 'USA' column name)
                if 'USA' in df.columns:
                     df_slice = df[['USA']]
            df_slice.columns = [sheet_name]
            dfs.append(df_slice)

        data = pd.concat(dfs, axis=1)
        data = data.astype(float)

        data.rename(columns={
            'MKT': 'Mkt-RF',     # TODO: double check!! This could just be mkt.
            'HML Devil': 'HML_Devil',
            'SMB': 'SMB_AQR', #
            #'RF': to RF_AQR or RF_{countrycode} when added 
        }, inplace=True)

        # Drops all data where HML_Devil is NaN
        #data = data.dropna(subset=['HML_Devil'])
        data = data.dropna(subset=['RF', 'UMD', 'HML_Devil'], how='any')
        data = data.round(8)
        return _process(data, self.start_date, self.end_date, filepath=self.output_file)
        # From func - cache stuff -- OLD CACHE STUFF
        #self.cache_dir = Path('~/.cache/getfactormodels/aqr/hml_devil').expanduser()
        #self.cache_dir.mkdir(parents=True, exist_ok=True)
        #self.cache = dc.Cache(str(self.cache_dir)) # diskcache requires a string path
# TODO: can see the last modified date in debug log; possibly set cache
# according to this



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
import pandas as pd
from dateutil import parser
from getfactormodels.utils.cli import parse_args
from getfactormodels.utils.utils import _get_model_key, _process, _save_to_file
from .models import (  # todo: dont import all models, just QFactor and FamaFrench (function needs changing)
    BarillasShankenFactors, CarhartFactors, DHSFactors, FamaFrenchFactors,
    HMLDevilFactors, ICRFactors, LiquidityFactors, MispricingFactors, QFactors)


# TEMPORARY MINIMAL REWORK (until the keymaps and insane regex is dropped
# and base class and FactorExtractor done)
# just getting get_factors to work!
def get_factors(model: str | int = 3,
                frequency: str = "m",
                start_date: str | None = None,
                end_date: str | None = None,
                output_file: str | None = None,
                *,
                region: str | None = None,
                ):
    """Get data for a specific factor model.

    Main public function.

    Args:
        model (str, float): the name of the factor model ('3', '4', '5', '6', 'carhart', 
                            'liq', 'misp', 'icr', 'dhs', 'qclassic', 'q', 'hml_d').
        frequency  (str, optional): the frequency of the data. ('d' 'w' 'm' 'q' or 'y', default: 'm')
        start_date (str, optional): start date of the returned data, YYYY-MM-DD.
        end_date (str, optional): end date of returned data, YYYY-MM-DD.
        output(str, optional): filepath to write returned data to, e.g. "~/some/dir/some_file.csv"

    """
    # NOTE: Parses input, assigns it to a factor model class, calls it, and returns the data.
    #
    # TODO: FIXME. kwargs. TODO: default outputs in CLI.
    #
    frequency = frequency.lower()
    region = region
    model_key = _get_model_key(model)

    factor_instance = None
    if region is not None and model_key not in ['3', '4', '5', '6']:
        _model = _get_model_key(model)
        raise ValueError(
            f"Region '{region}' is not supported for the '{_model}'. "
            "The region parameter is only available for Fama-French models (3, 4, 5, 6).",
        )
    if model_key in ["3", "4", "5", "6"]:
        factor_instance = FamaFrenchFactors(model=model_key,
                                            frequency=frequency,
                                            start_date=start_date,
                                            end_date=end_date,
                                            output_file=output_file,
                                            region=region)

    elif model_key == "Qclassic":
        factor_instance = QFactors(frequency=frequency, start_date=start_date,
                                   end_date=end_date, output_file=output_file, classic=True)

    else:
        # Class loading: tries CamelCaseFactors then UPPERCASEFactors
        class_name_camel = f"{model_key}Factors"
        class_name_upper = f"{model_key.upper()}Factors"

        # search for the class in the global scope
        FactorClass = globals().get(class_name_camel) or globals().get(class_name_upper)

        if not FactorClass:
            raise ValueError(f"Invalid model: '{model_key}'. "
                             f"Tried:'{class_name_camel}','{class_name_upper}'.")

        factor_instance = FactorClass(frequency=frequency,
                                      start_date=start_date,
                                      end_date=end_date,
                                      cache_ttl=86400,
                                      output_file=output_file)

    return factor_instance.download()

# CLI (TODO)
### zzzzzzzzzzzzzz. Old mess. Being removed. TODO: (one day) extract factors into a composite model.
class FactorExtractor:
    def __init__(self,
                 model: str = '3',
                 frequency: str = 'm',
                 start_date: str | None = None,
                 end_date: str | None = None,
                 output: str | None = None,
                 *,
                 region: str | None = None):
        self.model: str = model
        self.frequency: str = frequency
        self.start_date = self.validate_date_format(start_date) if start_date else None
        self.end_date = self.validate_date_format(end_date) if end_date else None
        self.output = output
        self._no_rf = False
        self._no_mkt = False
        self.df = None
        self.region = region

    def no_rf(self) -> None:
        self._no_rf = True

    def no_mkt(self) -> None:
        self._no_mkt = True

    def extract(self, data, factor: str | list[str]) -> pd.Series | pd.DataFrame:
        """Retrieves a single factor (column) from the dataset."""
        #data = self.download()

        if data.empty:
            raise RuntimeError("DataFrame empty: can not extract a factor.")

        if isinstance(factor, str):
            if factor not in data.columns:
                raise ValueError(f"Factor '{factor}' not available.")       
            return data[factor]

        elif isinstance(factor, list):
            return data[factor]

    @staticmethod
    def validate_date_format(date_string: str) -> str:
        """Validate the date format.
        Raises:
            ValueError: If the date format is incorrect.
        """
        try:
            return parser.parse(date_string).strftime("%Y-%m-%d")
        except ValueError as err:
            error_message = "Incorrect date format, use YYYY-MM-DD."
            raise ValueError(error_message) from err

    # Holy. TODO: Clean it up. Incorp into base 
    def get_factors(self) -> pd.DataFrame:
        """Fetch the factor data and store it in the class."""
        self.df = get_factors(
            model=self.model,
            frequency=self.frequency,
            start_date=self.start_date,
            end_date=self.end_date,
            output_file=self.output,  
            region=self.region)

        if self._no_rf:
            self.df = self.drop_rf(self.df.copy())  # create a copy before drop
        if self._no_mkt:
            self.df = self.drop_mkt(self.df.copy())

        return self.df

    def drop_rf(self, df: pd.DataFrame) -> pd.DataFrame:
        """Drop the ``RF`` column from the DataFrame."""

        if "RF" in df.columns:
            df = df.drop(columns=["RF"])
        else:
            print("`drop_rf` was called but no RF column was found.")

        return df

    def drop_mkt(self, df: pd.DataFrame) -> pd.DataFrame:
        """Drop the ``MKT`` column from the DataFrame."""

        if "Mkt-RF" in df.columns:
            df = df.drop(columns=["Mkt-RF"])
        else:
            print("`drop_mkt` was called but no MKT column was found.")

        return df

    def to_file(self, filename: str):
        """
        Save the factor data to a file.

        Args:
            filename (str): The name of the file to save the data to.
        """
        if self.df is None:
            raise ValueError("No data to save. Fetch factors first.")

        # TODO: could call _save_to_file directly
        _process(self.df, filepath=filename)  #lol FIXME


def main():
    args = parse_args()

    extractor = FactorExtractor(model=args.model,
                                frequency=args.frequency,
                                start_date=args.start,
                                end_date=args.end,
                                region=args.region,)
    if args.norf:
        extractor.no_rf()
    if args.nomktrf:
        extractor.no_mkt()

    df = extractor.get_factors()

    if args.extract:
        df = extractor.extract(df, args.extract)
    if args.output:
        output_path = Path(args.output).expanduser()

        if args.output:     # TODO: cache
            output_path = Path(args.output)
            min_date = df.index.min() 
            max_date = df.index.max()

            actual_start = min_date.strftime('%Y%m%d')
            actual_end = max_date.strftime('%Y%m%d')
            # if NaT, might be outside available data... to do warn/catch 

            _filename = f"{args.model}_{args.frequency.upper()}_{actual_start}-{actual_end}"
            _ext = '.csv'

            user_path = Path(args.output).expanduser()
            output_path = None

            if user_path.is_dir():
                output_path = user_path / (_filename + _ext)
            elif user_path.suffix:
                output_path = user_path
            else: #add ext to user provided 'file' or 'dir/file'
                output_path = user_path.parent / (user_path.name + _ext)

            if output_path:
                _save_to_file(df, output_path)
    print(df)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
from typing import Any
import pandas as pd
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.utils import (_pd_rearrange_cols, _save_to_file,
                                         _slice_dates)


# TODO: proper class docstr's.
class MispricingFactors(FactorModel):
    """Mispricing factors of Stambaugh & Yuan (2016).

    Downloads the Mispricing factors of R. F. Stambaugh and Y. Yuan. Data from
    1963 to 2016. Note: the SMB factor is returned as SMB_SY.

    Args:
        frequency(str, optional): 'm' (monthly), 'd' (daily)
        start_date (str, optional): The start date YYYY-MM-DD.
        end_date (str, optional): The end date YYYY-MM-DD.
        output_file (str, optional): Optional file path to save to file. Supports csv, pkl.
        classic (bool, optional): returns the classic 4-factor q-factor model. Default: False.
        cache_ttl (int, optional): Cached download time-to-live in seconds (default: 86400).
    
    Returns:
        pd.DataFrame: timeseries of factor data.

    References:
    - Pub: R. F. Stambaugh and Y. Yuan, ‘Mispricing Factors’, The Review of 
    Financial Studies, vol. 30, no. 4, pp. 1270–1315, 12 2016.

    Data source: https://finance.wharton.upenn.edu/~stambaug/
    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "m"]
    
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _get_url(self) -> str:
        base_url = "https://finance.wharton.upenn.edu/~stambaug"
        file_name = "M4d" if self.frequency == "d" else "M4"
        return f"{base_url}/{file_name}.csv"

    SCHEMA = pa.schema([
        ('YYYYMM', pa.timestamp('ms')),
        ('MKTRF', pa.float64()),
        ('SMB', pa.float64()),
        ('RMW', pa.float64()),
        ('CMA', pa.float64()),
        ('RF', pa.float64()),
    ])    

    def _read(self, data):
        """Reads the Mispricing factors CSV."""
        try:
            # TEST: reading CSV with PyArrow
            # will be a CSV reader
            read_opts = pv.ReadOptions(
                skip_rows=1,
                column_names=['YYYYMM', 'MKTRF', 'SMB', 'MGMT', 'PERF', 'RF'],
            )

            parse_opts = pv.ParseOptions(
                delimiter=',',
                ignore_empty_lines=True,
            )

            convert_opts = pv.ConvertOptions(
                column_types=self.SCHEMA,
                timestamp_parsers=["%Y%m%d", "%Y%m"],
            )

            table = pv.read_csv(
                io.BytesIO(data),
                read_options=read_opts,
                parse_options=parse_opts,
                convert_options=convert_opts,
            )

            table = table.rename_columns([
                'date',     # YYYYMM -> date
                'Mkt-RF',   # MKTRF -> Mkt-RF
                'SMB_SY',   # SMB -> SMB_SY
                'MGMT', 
                'PERF',
                'RF',
            ])

            # Debug, testing
            initial_rows = table.num_rows
            table = table.drop_null()
            final_rows = table.num_rows
            rows_dropped = initial_rows - final_rows

            if rows_dropped > 0:
                msg = (f"{rows_dropped} NaN rows dropped."
                      f"({initial_rows} -> {final_rows}).")
                self.log.debug(msg)
            else:
                msg = f"No NaNs in {final_rows} rows."
                self.log.debug(msg)

            # pandas line --------------------------------------------------- #
            df = table.to_pandas()

            df = df.set_index('date')
            df.index.name = 'date'

            #df = df.replace([-99.99, -999], pd.NA).dropna()

            if self.frequency == "m":
                # PyArrow gives the 1st of the month, shift to end.
                df.index = df.index + pd.offsets.MonthEnd(0)  # this is whats messing user input hmm
                # will keep uncommented while every other model does this.
            # TODO: If using end of month (which it is), then need to 
            # parse user input dates, THEN shift to end of month. FIXME.
            df = _pd_rearrange_cols(df)
            
            df = _slice_dates(df, self.start_date, self.end_date)

            if self.output_file:
                _save_to_file(df, filepath=self.output_file)

            return df
 
        # returns an empty dataframe that base class expects for this model.
        # TODO: FIXIME: cleanup, handle errors properly everywhere...
        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            err_msg = f"Reading or parsing failed for Mispricing factors: {e}"
            self.log.error(err_msg)
            column_names = ['Mkt-RF', 'SMB_SY', 'MGMT', 'PERF', 'RF']

            return pd.DataFrame(columns=column_names, # type: ignore [reportArgumentType] 
                                index=pd.DatetimeIndex([], name='date'))



# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
import zipfile
from typing import Any
import pandas as pd
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.http_client import HttpClient
from getfactormodels.utils.utils import _slice_dates

# TODO: yearly data isn't inclusive of end yr
#  but start_date includes partial start year/month !!
# eg start_date=1968-12-25, end_date=1970-01-01 with frequency='y' 
## will return yearly data for all of 1968 and none of 1970.
#
# Potentially warn user if month and day is passed to year? 
# Warn the returned data includes the full year, or not the last year because they specifed date...
#  just read YYYY?


class FamaFrenchFactors(FactorModel):
    # will do proper docstr later. TODO: utils and getting every model with pyarrow and pushing pandas to
    # boundaries.
    """Download Fama-French (and Carhart) factor models.

    Downloads the 3-factor (1993), 5-factor (2015), or 6-factor (2018)
    model of Fama & French, or Carhart's (1997) 4-factor model. Data
    is available in daily, weekly, monthly, and annual frequencies.
    Saves data to file if `output_file` is specified.

    Args:
        `model (str | int)`: model to return. `3`, `4`, `5` or `6` (default: 3)
        `frequency` (`str`): the frequency of the data. `d`, `m` `y` or `w`
        (default: `m`). Only the 3-Factor model is available for weekly freq.
        `start_date` (`str, optional`): the start date of the data, as 
            YYYY-MM-DD.
        `end_date` (`str, optional`): the end date of the data, as YYYY-MM-DD.
        `cache_ttl` (`int, optional`): Cached download time-to-live in seconds 
            (default: `86400`).
        `region` (`str, optional`): returns the region (intl/developed/emerging markets)
         model. New. Testing.

    Returns:
        `pd.Dataframe`: time series of returned data.

    References:
    - E. F. Fama and K. R. French, ‘Common risk factors in the returns on stocks 
    and bonds’, Journal of Financial Economics, vol. 33, no. 1, pp. 3–56, 1993.
    - E. F. Fama and K. R. French, ‘A five-factor asset pricing model’, Journal 
    of Financial Economics, vol. 116, no. 1, pp. 1–22, 2015.
    - E. F. Fama and K. R. French, ‘Choosing factors’, Journal of Financial 
    Economics, vol. 128, no. 2, pp. 234–252, 2018.
    """
    @property
    def _frequencies(self) -> list[str]:
        return ['d', 'w', 'm', 'y']

    FF_REGION_MAP = {
        'us': 'us',
        'emerging': 'Emerging',
        'developed': 'Developed',
        'developed ex us': 'Developed_ex_US',
        'ex us': 'Developed_ex_US',
        'ex-us': 'Developed_ex_US',
        'europe': 'Europe',
        'japan': 'Japan',
        'asia pacific ex japan': 'Asia_Pacific_ex_Japan',
        'north_america': 'North_America',
        'na': 'North_America',
    } # TODO: Clean this up, and region map usage throughout (after factorextractor gets removed)

    def __init__(self,
                 frequency: str = 'm',
                 model: int|str = '3',
                 region: str | None = 'us',
                 **kwargs: Any) -> None:
        self.frequency = frequency.lower()
        self.model = str(model)
        # eg: japan -> Japan [FIXME, lower everywhere]
        self.region = None if region is None else self.FF_REGION_MAP.get(region.lower(), None)
        self._validate_ff_input()

        super().__init__(frequency=frequency, model=model, **kwargs)

    def _validate_ff_input(self):
        if self.model not in ["3", "4", "5", "6"]:
            raise ValueError(
                f"Invalid model '{self.model}': must be '3' '4' '5' or '6'",
            )
        if self.frequency == 'w' and self.model not in {"3"}: #no 4 weekly mom
            raise ValueError(f"Fama French weekly data is only available for the 3 Factor model ({self.model})")

        if self.frequency != 'm' and self.region == "Emerging":
            raise ValueError(
                "Emerging Markets data is only available in monthly frequency.",
            )
        if self.frequency not in ['d', 'm', 'y'] and self.region not in ['US', 'us', None]:
            raise ValueError(
                "Region data is only available in daily, monthly and yearly frequency.",
            )       
        # self.region is None if invalid region, or it is mapped! (e.g., 'Japan', 'us').
        valid_mapped_regions = set(self.FF_REGION_MAP.values()) | {None}

        if self.region not in valid_mapped_regions:
            valid_region_keys = ', '.join(f'`{k}`' for k in self.FF_REGION_MAP)
            raise ValueError(
                f"Invalid region. Must be one of: {valid_region_keys}",
            )


    def _get_url(self) -> str:
        """Constructs the URL for downloading Fama-French data based on region, model, and frequency."""
        base_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"
        region = self.region
        model = self.model
        frequency = self.frequency.lower()

        file_name = None
        basefilename = None # Initialize basefilename here for clarity

        if region in ['us', None]:
            if model in {"3", "4"}:
                basefilename = "F-F_Research_Data_Factors"
            elif model in {"5", "6"}:
                basefilename = "F-F_Research_Data_5_Factors_2x3"
            else:
                raise ValueError(f"Invalid model '{model}' for US factors. Must be 3, 4, 5, or 6.")

            freq_suffix = ""
            if frequency == 'd':
                freq_suffix = "_daily"
            elif frequency == 'w':
                freq_suffix = "_weekly"

            file_name = f"{basefilename}{freq_suffix}_CSV.zip"
            self.log.debug(f"filename: {file_name}")
            self.log.debug(f"full url: {base_url}/{file_name}")
            return f"{base_url}/{file_name}"

        if model in ['3', '4', '5', '6']:
            base_ff_model = '3' if model in ['3', '4'] else '5'
        else:
            raise ValueError(f"Invalid model '{model}' for regional factors in region '{region}'.")
        freq_suffix = '_Daily' if frequency == 'd' else ''

        if region == 'Emerging' and frequency == 'm':
            file_name = 'Emerging_5_Factors_CSV.zip'
        elif region is not None:
            file_name = f'{region}_{base_ff_model}_Factors{freq_suffix}_CSV.zip'

        return f'{base_url}/{file_name}'

    def _get_mom_url(self, frequency, region):
        """Constructs the URL for momentum factors."""
        base_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"

        if region not in ["us", None]: # Note: region is mapped 'Emerging', not 'emerging'
            if self.model in ['4', '6']:
                freq_suffix = "_Daily" if frequency == 'd' else ""
                file_name = f'{region}_Mom_Factor{freq_suffix}_CSV.zip' 
                return f"{base_url}/{file_name}"

       # if region == 'Emerging':
       #     file = "Emerging_MOM_Factor_TXT.zip" #only m; forgot why TXT 

        elif frequency == 'd':
            file = "F-F_Momentum_Factor_daily_CSV.zip"

        else:
            file = "F-F_Momentum_Factor_CSV.zip"

        return f"{base_url}/{file}"

    # ------------------------------------------------------------------------- #

    @staticmethod
    def _read_zip(_data, frequency, region):
        """Download and read CSV from zip file."""
        # old function stuff....
        
        try:
            with zipfile.ZipFile(io.BytesIO(_data)) as zip_file:
                filename = zip_file.namelist()[0]

                with zip_file.open(filename) as file:
                    content = file.read().decode('utf-8')
                if region == 'Emerging':
                    skip_rows = 6
                else:
                    skip_rows = 12 if 'momentum' in filename.lower() else 4

                # Monthly files contain the annual data, under this line
                annual_marker = " Annual Factors: January-December"
                marker_pos = content.find(annual_marker)
                
                if frequency in ['y', 'm'] and marker_pos != -1:
                    if frequency == 'm':
                        content = content[:marker_pos]
                    else:  # freq == 'y'
                        all_lines = content[marker_pos:].split('\n')
                        if len(all_lines) > 2:
                            # Join lines starting from the Header Line
                            content = '\n'.join(all_lines[1:])
                            skip_rows = 0

                df = pd.read_csv(
                    io.StringIO(content),
                    skiprows=skip_rows,
                    index_col=[0],
                    header=0,
                    parse_dates=False,
                    skipfooter=1,
                    engine="python",
                    na_values=[-99.99, '-99.99'],  #stated in emerging, momentum; every model? 
                    na_filter=True,   #fix to mom data  # NANS ARE RETURNING AS -0.99999 coz decimalizing
                    on_bad_lines='error')

            df.index = df.index.astype(str)
            df.index = df.index.str.strip()
            df.index.name = "date"
            #df = df.dropna(how='all')
            return df

        except Exception as e:
            print(f"Error parsing zip file: {e}")
            return pd.DataFrame()

   
    # parse_date=True prob handles this now: keeping for now.   #TODO: remove
    def _parse_dates(self, df): #TODO: types everywhere.
        """Parse the date index based on frequency."""
        if df.empty:
            return df

        if self.frequency == 'm':
            # Monthly data YYYYMM (6 chars)
            mask = df.index.str.len() == 6
            df = df[mask]
            if not df.empty:
                df.index = pd.to_datetime(df.index, format='%Y%m') + pd.offsets.MonthEnd(0)

        elif self.frequency == 'y':
            mask = df.index.str.len() == 4
            df = df[mask]
            if not df.empty:
                df.index = pd.to_datetime(df.index, format='%Y') + pd.offsets.YearEnd(month=12)

        elif self.frequency in {'d', 'w'}:
            # Day/week data shares YYYYMMDD format
            mask = df.index.str.len() == 8
            df = df[mask]
            if not df.empty:
                df.index = pd.to_datetime(df.index, format='%Y%m%d')

        df.index.name = "date"
        return df


    def _download_mom_data(self, frequency):
        url = self._get_mom_url(frequency, self.region)
        try:
            with HttpClient(timeout=10.0) as client:
                _zip = client.download(url)
            return self._read_zip(_zip, frequency, self.region)
        except Exception as e: #todo exception
            raise RuntimeError(f"Failed to download momentum data from {url}.") from e


    def _add_momentum_factor(self, df):
        if self.model not in {"4", "6"}:
            return df 
        
        try:
            mom_data = self._download_mom_data(self.frequency)
        
        except RuntimeError as e:
            self.log.warning(f"Momentum factor download failure: {e}")
            return df
        
        if mom_data.empty:
            print("Empty df.")
            return df
        
        mom_data = self._parse_dates(mom_data)

        if self.region not in ['us', None]:
            mom_name = "WML"  # winner minus loser in the EM and intl data
        elif self.model == "6":    # 6 factor call it Up Minus Down
            mom_name = "UMD"
        elif self.model == "4":   # and MOM in 4-factor (TODO: name MOM for just MOM?)
            mom_name = "MOM"
 
        if len(mom_data.columns) > 0:    
            mom_series = (
                mom_data.iloc[:, 0]  # Extracts first col,
                .rename(mom_name) # Renames it
                .reindex(df.index)
            )

            df = df.join(mom_series)
            self.log.info(f"Added momentum factor: {mom_name}")

            if self.region == 'us':
                df = df.dropna(how='any')  # Drop NaNs after MOM added; trims models to MOM size 
            else:
                # eg, emerging RMW/CMA begins with NaNs -- keepin
                # eg, asia pacific ex japan starts with 4 months of NaNs in WML (4 in monthly) -- droppin
                # but all factors go back to 1990-07-02. 
                #If momentum model, then we drop the nan WML rows...
                if self.model in ['4', '6']:
                    df = df.dropna(how='any')
                else:
                    df = df.dropna(how='all')  # in dev/emerging/intl markets the mom factor if often shorter than the dataset. (But if models are momentum models... might just trim to size as well)
        else:        
            self.log.warning("Mom data contained no columns after parsing...")
        return df 


    def _read(self, data) -> pd.DataFrame:
        try:
            df = self._read_zip(data, self.frequency, self.region)

            if df.empty:
                print("Warning: No data downloaded")
                return pd.DataFrame()

            # Parse dates and add momentum if needed
            df = self._parse_dates(df)
            df = self._add_momentum_factor(df)
            df = _slice_dates(df, self.start_date, self.end_date)
            df = df.sort_index()
            # Just for now until rewrite... later.
            if self.region == 'Emerging':
                if self.model == '3':
                    df = df.get(["Mkt-RF", "SMB", "HML", "RF"], df)
                if self.model == '4':
                    df = df.get(["Mkt-RF", "SMB", "HML", "WML", "RF"], df)
                if self.model == '5':
                    df = df.get(["Mkt-RF", "SMB", "HML", "RMW", "CMA", "RF"], df)
                #if 6 then it's all here...

            # Decimalize. Keep here for now. Finicky.
            df = df.astype(float)
            return df * 0.01
            # TODO: drop nans/nulls, after adding MOM, and for barillas shanken, etc.

        except Exception as e:
            print(f"Error downloading Fama-French data: {e}")
            raise

# Welllll its using the base class at least... TODO FIXME FIXME 



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse


def parse_args() -> argparse.Namespace:
    """Argument parser, allowing for command line arguments.
    This is the function used in pyproject.toml to run the CLI."""
    parser = argparse.ArgumentParser(
        description='Download datasets for various factor models.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='''  Example usage:
            getfactormodels --model ff3 --frequency m
            getfactormodels --model liquidity --frequency m --start 2000-01-01 --end 2009-12-31
            getfactormodels -m 5 -f m -e 2009-12-31 --extract SMB RF -o '~/file.csv'
            getfactormodels -m Carhart -f d -s 2000-01-01 -e 2009-12-31 -x MOM -o file 

            '''
    )
    parser.add_argument('-m', '--model', type=str, required=True, metavar="MODEL_ID",
                        help="the model to use. eg, 'ff3', 'liquidity', 'dhs', 'q', etc. "
                        "Accepts 3, 4, 5, 6 in place of ff3, carhart, ff5, ff6.")
    parser.add_argument('-f', '--frequency', type=str, required=False, default='m',
                        choices=['d', 'w', 'm', 'q', 'y'], metavar="INTERVAL",
                        help="Frequency of the data. 'd', 'w', 'm', 'q', 'y'.")
    parser.add_argument('-s', '--start', type=str, required=False, metavar="YYYY-MM-DD",
                        help='The start date for the data.')
    parser.add_argument('-e', '--end', type=str, required=False, metavar="YYYY-MM-DD",
                        help='The end date for the data.')
    parser.add_argument('-o', '--output', type=str, required=False, default=None, metavar="FILEPATH",
                        help='The file to save the data to.')
    parser.add_argument('-x', '--extract', required=False, nargs='+', metavar="FACTOR_NAME",
                        help='Extract specific factor(s) from a model, e.g., "SMB".')
    parser.add_argument('-r', '--region', type=str, required=False, metavar="REGION_ID",
                        choices=['us', 'developed', 'developed ex us', 'europe', 'japan',
                                 'asia pacific ex japan', 'north america', 'emerging'],
                        help="Fama-French models only: get data for devleoped and emerging markets. "
                        "One of: 'us', 'developed', 'developed ex us', 'europe', 'japan',"
                        " 'asia pacific ex japan', 'north america', 'emerging'.")
    # Being removed, adding a drop functionality:
    parser.add_argument('-R', '--norf', action='store_true', help='Don\'t include the risk-free rate column (RF).')
    parser.add_argument('-M', '--nomktrf', action='store_true', help='Drop the Mkt-RF column.')

    return parser.parse_args()



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import hashlib
import logging
import ssl
import certifi
import httpx
from platformdirs import user_cache_path
from .cache import _Cache

log = logging.getLogger(__name__)
# TODO: errors

class HttpClient:
    """Simple http client: wrapper around httpx.Client with caching."""
    
    APP_NAME = "getfactormodels"  #platformdirs for xdg cache 

    def __init__(self, timeout: float | int = 15.0,
                 cache_dir: str | None = None, # None by default!
                 default_cache_ttl: int = 86400): # 1 day default
        self.timeout = timeout
        self.default_cache_ttl = default_cache_ttl
        self._client = None # initializes client as None

        # XDG path
        if cache_dir is None:
            _cache_path = user_cache_path(appname=self.APP_NAME, 
                                       ensure_exists=True)
            self.cache_dir = str(_cache_path.resolve())
            
        else:
            # user explicitly passed a path, use it...
            self.cache_dir = cache_dir

        msg = f"cache directory: {self.cache_dir}"
        log.debug(msg)

        self.cache = _Cache(self.cache_dir, default_timeout=default_cache_ttl)

    def __enter__(self):
        if self._client is None:
            # fix: tox depre warning: create SSL context
            ssl_context = ssl.create_default_context(cafile=certifi.where())
            self._client = httpx.Client(
                verify=ssl_context,
                timeout=self.timeout,
                follow_redirects=True,
                max_redirects=3,
            )

        return self 
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        msg = f"closed: {self._client.__class__.__name__}"
        log.debug(msg)

    def close(self) -> None:
        if self._client is not None:
            msg = f'closing {self._client.__class__.__name__}'
            log.debug(msg)  # No print in log messages, ruff
            self._client.close()
            self._client = None

        msg =f'closing {self.cache.__class__.__name__}'
        log.debug(msg)
        self.cache.close()

    def _generate_cache_key(self, url: str) -> str:
        """Generates a cache key/hash for the URL."""
        return hashlib.sha256(url.encode('utf-8')).hexdigest()

    def download(self, url: str, cache_ttl: int | None = None) -> bytes:
        """Downloads content from a given URL.
           cache_ttl: int, secs
        """
        # To prevent non-context manager usage?
        if self._client is None:
            raise ClientNotOpenError(
                "HttpClient is not open. It must be used within a 'with HttpClient(...) as client:' block.",
            )

        cache_key = self._generate_cache_key(url)
        cached_data = self.cache.get(cache_key)

        if cached_data is not None:
            return cached_data

        try:
            msg = f'Connecting: {url[:30]}...'
            log.info(msg)

            response = self._client.get(url)
            response.raise_for_status()
            data = response.content  # Always bytes

            # Store in cache: need to verify it
            self.cache.set(cache_key, data, expire_secs=cache_ttl)
            msg = f"{len(data)} bytes cached."
            log.info(msg)

            return data
        except httpx.HTTPStatusError as e:
            err = f"HTTP error {e.response.status_code} for {url}"
            log.error(err)
            raise ConnectionError(f"HTTP error: {e.response.status_code}") from e
        except httpx.RequestError as e:
            err = f"Network error for {url}: {e}"
            log.error(err)
            raise ConnectionError(f"Request error: {e}") from e

    def check_connection(self, url: str):
        """Simple url ping. Boolean."""
        check_timeout = 4.0

        if self._client is None:
            raise ClientNotOpenError("HttpClient is not open. Use in a `with` block.")

        try:
            msg = f"Attempting HEAD: {url}..."
            log.info(msg)

            response = self._client.head(url, timeout=check_timeout)

            if response.is_success:
                msg = f"URL:{url}\nstatus: {response.status_code}"
                log.info(msg)
                return True

            msg = "Falling back to GET..."
            log.info(msg)

            response = self._client.get(url, timeout=check_timeout)

            if response.is_success:
                return True

            msg = "Couldn't establish connection."
            log.info(msg)
        
        except httpx.RequestError: 
            return False

        return False

    # TODO: user needs to acces this. force, or clear cache?
    def _clear_cache(self) -> None:
        self.cache.clear()
        log.debug("cache cleared by HttpClient")

# TODO: Exception handling...
class ClientNotOpenError(Exception):
    """Raised when HttpClient is used outside of a 'with' block."""
    pass



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
from .main import FactorExtractor, get_factors
from .models import (BarillasShankenFactors, CarhartFactors, DHSFactors,
                     FamaFrenchFactors, HMLDevilFactors, ICRFactors,
                     LiquidityFactors, MispricingFactors, QFactors)

logging.getLogger('getfactormodels').setLevel(logging.ERROR)

__version__ = "0.0.9"
__all__ = [ "MispricingFactors",
           "DHSFactors",
           "ICRFactors",
           "QFactors",
           "LiquidityFactors",
           "CarhartFactors",
           "HMLDevilFactors",
           "FamaFrenchFactors",
           "BarillasShankenFactors",
           "FactorExtractor",
           "get_factors",
           ]




#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, override
import pandas as pd
from getfactormodels.utils.utils import _slice_dates
from .base import FactorModel
from .fama_french import FamaFrenchFactors
from .hml_devil import HMLDevilFactors
from .q_factors import QFactors


class BarillasShankenFactors(FactorModel):
    """Download the Barillas-Shanken 6 Factor Model (2018).

    A combination of the 5-factor model of Fama and French (2015), the q-factor
    model of Hou, Xue, and Zhang (2015), and Asness and Frazzini's HML Devil (2013).
    This is the factor model with the highest posterior inclusion probability
    in Barillas and Shanken (2018).

    Args:
        `frequency` (`str`): the frequency of the data. `d`, `m` (default: `m`)
        (default: `m`). Note: Only the 3 factor model is available in weekly.
        `start_date` (`str, optional`): the start date of the data, as 
            YYYY-MM-DD.
        `end_date` (`str, optional`): the end date of the data, as YYYY-MM-DD.
        `cache_ttl` (`int, optional`): Cached download time-to-live in seconds 
            (default: `86400`).

    Note:
        - Relies on the HML Devil factors being retrieved (which is very slow).

    Returns:
        pd.DataFrame: A timeseries of the factor data.

    References:
    - F. Barillas and J. Shanken, ‘Comparing Asset Pricing Models’, Journal of 
    Finance, vol. 73, no. 2, pp. 715–754, 2018.
    """
    # Not ideal but working
    #TODO: extract instead of dl
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "m"]

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _get_url(self) -> str:  # I guess?
        msg = "_read called on BarillasShanken. Constructed from other models. No URL."
        self.log.warning(msg)
        return ""

    def _read(self, data: bytes) -> pd.DataFrame:
        # Must exist to satisfy the base class contract.
        self.log.warning("_read called on a constructed model (Barillas Shanken). Don't do that.")
        return pd.DataFrame()

    @override
    def download(self) -> pd.DataFrame:
        """Overrides base download() for barillas shanken."""
        if self._data is not None:
            self.log.debug("Data loaded. Returning stored DataFrame.")
            return self._data

        df = self._construct_bs()
        
        self._data = df 
        
        return df
    # ----------------------------------------------------------------- #
    def _construct_bs(self) -> pd.DataFrame:
        """Constructs the Barillas 6 factor model from other models"""

        print("  - Getting q factors...")
        _q = QFactors(frequency=self.frequency, 
                      classic=True).extract(['R_IA', 'R_ROE']) 
        # test: if this fails .extract is bad
        
        print("  - Getting Fama-French factors...")
        ffdata = FamaFrenchFactors(model='6', frequency=self.frequency)
        ff = ffdata.download()[['Mkt-RF', 'SMB', 'UMD']]

        # Merge q and Fama-French factors
        df = _q.merge(ff, left_index=True,
                     right_index=True, how='inner')

        print("  - Getting HML_Devil factor (this can take a while,"
            "please be patient)...")

        hmld_data = HMLDevilFactors(frequency=self.frequency,
                                    start_date=self.start_date,
                                    end_date=self.end_date)

        hml_d = hmld_data.download()
        # NOTE: Taking the 'RF' from AQR's series since it's here,
        # and it's the same data as Fama-French but to 4 decimals. 
        # Mkt-RF shows a difference though -- and bs6 should use 
        # the mkt-rf of ff!
        hml_d = hml_d[['HML_Devil', 'RF']]

        hml_d.index.name = 'date'
        df = df.merge(hml_d, left_index=True, right_index=True, how='inner')
        df = _slice_dates(df, self.start_date, self.end_date)

        return df



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import re
from datetime import datetime
from pathlib import Path
from types import MappingProxyType
import pandas as pd

log = logging.getLogger(__name__) #TODO: consistent logging.

__model_input_map = MappingProxyType({
    "3": r"\b((f?)f)?3\b|(ff)?1993",
    "5": r"\b(ff)?5|ff2015\b",
    "4": r"\b(c(ar(hart)?)?4?|ff4|carhart1997|4)\b",
    "6": r"\b(ff)?6|ff2018\b",
    "Q": r"\b(q(5)?|hmxz)\b",
    "Qclassic": r"\b(q4|q(_)?classic)|classic_q\b",
    "Mispricing": r"\b(sy4?|mispricing)|misp|yuan$|m4|mis|sy\b",
    "Liquidity": r"^(il)?liq(uidity)?|(pastor|ps|sp)$",
    "ICR": r"\bicr|hkm\b",
    "DHS": r"^(\bdhs\b|behav.*)$",
    "HMLDevil": r"\bhml(_)?d(evil)?|hmld\b",
    "BarillasShanken": r"\b(bs|bs6|barillas|shanken)\b" })


def _get_model_key(model):
    """Convert a model name to a model key.

    >>> _get_model_key('ff1993')
    '3'
    >>> _get_model_key('liQ')
    'Liquidity'
    >>>  _get_model_key('q4_factors')
    'Qclassic'
    >>> _get_model_key('ICR')
    'ICR'
    """
    model = str(model)

    for key, regex in __model_input_map.items():
        if re.match(regex, model, re.I):
            return key
    raise ValueError(f'Invalid model: {model}')


# TODO: Will redo as a Writer class with use pyarrow
# changing: no longer uses filename, output_dir, just filepath. Always returns Path 
def _prepare_filepath(filepath=None) -> Path:
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
    if filepath is None:
        return Path.cwd() / f"data_{timestamp}.csv"

    user_path = Path(filepath).expanduser()

    if user_path.is_dir():
        final_path = user_path / f"data_{timestamp}.csv"
    else:
        user_path.parent.mkdir(parents=True, exist_ok=True)
        final_path = user_path

    return final_path


def _save_to_file(data, filepath=None):
    if not isinstance(data, (pd.DataFrame, pd.Series)):
            raise ValueError('Data is not a pandas DataFrame or Series')
    full_path = _prepare_filepath(filepath)
    
    if full_path is None:
        raise ValueError("Failed to prepare filepath")
    
    extension = full_path.suffix.lower()
 
    if full_path.is_file():
        print(f'File exists: {full_path.name} - overwriting...')

    try:
        if extension == '.txt':
            data.to_csv(str(full_path), sep='\t')
        elif extension == '.csv':
            data.to_csv(str(full_path))
        elif extension == '.pkl':
            data.to_pickle(str(full_path))
        else:
            supported = ['.txt', '.csv', '.pkl']  # to add: feather, parquet, json
            raise ValueError(f'Unsupported file extension: {extension}. Must be one of: {supported}')
        print(f"File saved to: {full_path}")
    except Exception as e:
        # Fix UP024: Use OSError instead of IOError
        # Fix B904: Use from e to chain the exception
        raise OSError(f"Failed to save file to {full_path}: {str(e)}") from e


# TODO: check if ICR model has no RF or Mkt Excess return column
def _pd_rearrange_cols(data):
    """Rearranges columns of a df to put 'Mkt-RF' first and 'RF' last."""
    if isinstance(data, pd.Series):
        return data

    if not isinstance(data, pd.DataFrame):
        raise ValueError("Input must be a pandas DataFrame or Series")

    cols = list(data.columns)
    
    if 'Mkt-RF' in cols:
        cols.insert(0, cols.pop(cols.index('Mkt-RF')))
        log.debug("`Mkt-RF` column moved to start")

    if 'RF' in cols:
        cols.append(cols.pop(cols.index('RF')))
        log.debug("`RF` column moved to end of df.")

    return data.loc[:, cols]


def _validate_date(date_input):
    """Converts date formats to a standardized str format."""
    if date_input is None:
        return None

    if isinstance(date_input, int):
        # int should be in the format YYYYMMDD or YYYYMM.
        date_str = str(date_input)
        
        if len(date_str) == 8: # YYYYMMDD
            return f"{date_str[:4]}-{date_str[4:6]}-{date_str[6:]}"
        
        elif len(date_str) == 6:
             return f"{date_str[:4]}-{date_str[4:]}-01"
             
        else:
             raise ValueError(f"invalid length int: {date_str}")
    
    # is a timestamp
    if isinstance(date_input, pd.Timestamp):
        return date_input.strftime("%Y-%m-%d")

    # is str input
    if isinstance(date_input, str):
        try:
            # cnvert to datetime
            return pd.to_datetime(date_input).strftime("%Y-%m-%d")
        except (ValueError, pd.errors.ParserError) as err:
            raise ValueError("Incorrect date format, use YYYY-MM-DD.") from err
    try:
        # already a dt object
        return date_input.strftime("%Y-%m-%d")
    except AttributeError:
        raise TypeError(f"Unsupported date type: {type(date_input)}")


def _slice_dates(data, start_date=None, end_date=None):
    """Slice the dataframe to the specified date range."""
    if start_date is None and end_date is None:
        return data

    if start_date is not None:
        start_date = _validate_date(start_date)
    if end_date is not None:
        end_date = _validate_date(end_date)

    start = _validate_date(start_date) if start_date else None
    end = _validate_date(end_date) if end_date else None

    if not isinstance(data.index, pd.DatetimeIndex):
        try:
            data.index = pd.to_datetime(data.index)
        except Exception as e:
            raise ValueError("error parsing dates") from e
    
    return data.loc[start:end]

# Change: moved filepath stuff to a _prepare_filepath helper for now...
def _process(data, start_date=None, end_date=None, filepath=None):
    if not isinstance(data, (pd.DataFrame, pd.Series)):
        raise ValueError("Input data must be a pandas DataFrame or Series")

    data = _pd_rearrange_cols(data)
    data = _slice_dates(data, start_date, end_date)

    if filepath:
        _save_to_file(data, filepath=filepath)

    return data
    

